import argparse
import importlib
import logging
import os
import pathlib
import sys

from . import j1939, logreader, tester
from .errors import Error
from .version import __version__

# Remove once less users are using the old package structure.
from . import database as db  # isort: skip

__author__ = 'Erik Moqvist'

class _ErrorSubparser:
    def __init__(self, subparser_name, error_message):
        self.subparser_name = subparser_name
        self.error_message = error_message

    def add_subparser(self, subparser_list):
        err_parser = \
            subparser_list.add_parser(self.subparser_name,
                                      description = self.error_message)
        err_parser.add_argument("args", nargs="*")

        err_parser.set_defaults(func=self._print_error)

    def _print_error(self, args):
        raise ImportError(self.error_message)

def _load_subparser(subparser_name, subparsers):
    """Load a subparser for a CLI command in a safe manner.

    i.e., if the subparser cannot be loaded due to an import error or
    similar, no exception is raised if another command was invoked on
    the CLI."""

    try:
        result = importlib.import_module(f'.subparsers.{subparser_name}',
                                         package='cantools')
        result.add_subparser(subparsers)

    except ImportError as e:
        result = _ErrorSubparser(subparser_name,
                                 f'Command "{subparser_name}" is unavailable: "{e}"')
        result.add_subparser(subparsers)

def _main():
    parser = argparse.ArgumentParser(
        description='Various CAN utilities.',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument('-d', '--debug', action='store_true')
    parser.add_argument('--version',
                        action='version',
                        version=__version__,
                        help='Print version information and exit.')

    # Workaround to make the subparser required in Python 3.
    subparsers = parser.add_subparsers(title='subcommands',
                                       dest='subcommand')
    subparsers.required = True


    # load all subparses which have a source file in the cantools
    # module's 'subparsers' sub-directory
    subparsers_dir = pathlib.Path(__file__).parent / 'subparsers'
    for cur_file_name in os.listdir(subparsers_dir):
        if cur_file_name.startswith('__'):
            continue

        if cur_file_name.endswith('.py'):
            subparser_name = cur_file_name[:-3]
            _load_subparser(subparser_name, subparsers)
        elif (subparsers_dir / cur_file_name / "__init__.py").is_file():
            subparser_name = cur_file_name
            _load_subparser(subparser_name, subparsers)

    args = parser.parse_args()

    if args.debug:
        logging.basicConfig(level=logging.DEBUG)
        args.func(args)
    else:
        try:
            args.func(args)
        except BaseException as e:
            sys.exit('error: ' + str(e))



import unittest
import io

import cantools


class TestLogreaderFormats(unittest.TestCase):
    def test_empty_line(self):
        parser = cantools.logreader.Parser()
        outp = parser.parse("")
        self.assertIsNone(outp)

    def test_candump(self):
        parser = cantools.logreader.Parser()

        outp = parser.parse("vcan0  0C8   [8]  F0 00 00 00 00 00 00 00")
        self.assertEqual(outp.channel, 'vcan0')
        self.assertEqual(outp.frame_id, 0xc8)
        self.assertEqual(outp.data, b'\xF0\x00\x00\x00\x00\x00\x00\x00')
        self.assertEqual(outp.timestamp_format, cantools.logreader.TimestampFormat.MISSING)

        outp = parser.parse("  vcan1  064   [10]  F0 01 FF FF FF FF FF FF FF FF")
        self.assertEqual(outp.channel, 'vcan1')
        self.assertEqual(outp.frame_id, 0x64)
        self.assertEqual(outp.data, b'\xF0\x01\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF')
        self.assertEqual(outp.timestamp_format, cantools.logreader.TimestampFormat.MISSING)

        outp = parser.parse("  vcan0  ERROR")
        self.assertIsNone(outp)

        outp = parser.parse("  vcan0  1F4   [4]  01 02 03 04")
        self.assertEqual(outp.channel, 'vcan0')
        self.assertEqual(outp.frame_id, 0x1f4)
        self.assertEqual(outp.data, b'\x01\x02\x03\x04')
        self.assertEqual(outp.timestamp_format, cantools.logreader.TimestampFormat.MISSING)

        outp = parser.parse("  vcan0  1F3   [3]  01 02 03")
        self.assertEqual(outp.channel, 'vcan0')
        self.assertEqual(outp.frame_id, 0x1f3)
        self.assertEqual(outp.data, b'\x01\x02\x03')
        self.assertEqual(outp.timestamp_format, cantools.logreader.TimestampFormat.MISSING)

    def test_timestamped_candump(self):
        parser = cantools.logreader.Parser()

        outp = parser.parse("(000.000000)  vcan0  0C8   [8]  F0 00 00 00 00 00 00 00")
        self.assertEqual(outp.channel, 'vcan0')
        self.assertEqual(outp.frame_id, 0xc8)
        self.assertEqual(outp.data, b'\xF0\x00\x00\x00\x00\x00\x00\x00')
        self.assertEqual(outp.timestamp_format, cantools.logreader.TimestampFormat.RELATIVE)
        self.assertEqual(outp.timestamp.seconds, 0.0)

        outp = parser.parse("(002.047817)  vcan0  064   [8]  F0 01 FF FF FF FF FF FF")
        self.assertEqual(outp.channel, 'vcan0')
        self.assertEqual(outp.frame_id, 0x64)
        self.assertEqual(outp.data, b'\xF0\x01\xFF\xFF\xFF\xFF\xFF\xFF')
        self.assertEqual(outp.timestamp_format, cantools.logreader.TimestampFormat.RELATIVE)
        self.assertEqual(outp.timestamp.seconds, 2)
        self.assertEqual(outp.timestamp.microseconds, 47817)

        outp = parser.parse("(012.831664)  vcan0  1F4   [4]  01 02 03 04")
        self.assertEqual(outp.channel, 'vcan0')
        self.assertEqual(outp.frame_id, 0x1f4)
        self.assertEqual(outp.data, b'\x01\x02\x03\x04')
        self.assertEqual(outp.timestamp_format, cantools.logreader.TimestampFormat.RELATIVE)
        self.assertEqual(outp.timestamp.seconds, 12)
        self.assertEqual(outp.timestamp.microseconds, 831664)

        outp = parser.parse("(015.679614)  vcan0  1F3   [3]  01 02 03")
        self.assertEqual(outp.channel, 'vcan0')
        self.assertEqual(outp.frame_id, 0x1f3)
        self.assertEqual(outp.data, b'\x01\x02\x03')
        self.assertEqual(outp.timestamp.seconds, 15)
        self.assertEqual(outp.timestamp.microseconds, 679614)
        self.assertEqual(outp.timestamp_format, cantools.logreader.TimestampFormat.RELATIVE)

        outp = parser.parse("(1613749650.388103)  can1       0AD  [08]  A6 55 3B CF 3F 1A F5 2A")
        self.assertEqual(outp.channel, 'can1')
        self.assertEqual(outp.frame_id, 0xad)
        self.assertEqual(outp.data, b'\xA6\x55\x3B\xCF\x3F\x1A\xF5\x2A')
        self.assertEqual(outp.timestamp_format, cantools.logreader.TimestampFormat.ABSOLUTE)
        self.assertEqual(outp.timestamp.year, 2021)
        self.assertEqual(outp.timestamp.month, 2)
        self.assertEqual(outp.timestamp.day, 19)
        self.assertEqual(outp.timestamp.hour, 15)
        self.assertEqual(outp.timestamp.minute, 47)
        self.assertEqual(outp.timestamp.second, 30)
        self.assertEqual(outp.timestamp.microsecond, 388103)

    def test_candump_log(self):
        parser = cantools.logreader.Parser()

        outp = parser.parse("(1579857014.345944) can2 486#82967A6B006B07F8")
        self.assertEqual(outp.channel, 'can2')
        self.assertEqual(outp.frame_id, 0x486)
        self.assertEqual(outp.data, b'\x82\x96\x7A\x6B\x00\x6B\x07\xF8')
        self.assertEqual(outp.timestamp_format, cantools.logreader.TimestampFormat.ABSOLUTE)
        self.assertEqual(outp.timestamp.year, 2020)
        self.assertEqual(outp.timestamp.month, 1)
        self.assertEqual(outp.timestamp.day, 24)
        self.assertEqual(outp.timestamp.hour, 9)
        self.assertEqual(outp.timestamp.minute, 10)
        self.assertEqual(outp.timestamp.second, 14)
        self.assertEqual(outp.timestamp.microsecond, 345944)

        outp = parser.parse(
            "(1613656104.501098) can3 14C##155B53476F7B82EEEB8E97236AC252B8BBB5B80A6A7734B2F675C6D2CEEC869D3")
        self.assertEqual(outp.channel, 'can3')
        self.assertEqual(outp.frame_id, 0x14c)
        self.assertEqual(
            outp.data, b'\x55\xB5\x34\x76\xF7\xB8\x2E\xEE\xB8\xE9\x72\x36\xAC\x25\x2B\x8B\xBB\x5B\x80\xA6\xA7\x73\x4B\x2F\x67\x5C\x6D\x2C\xEE\xC8\x69\xD3')
        self.assertEqual(outp.timestamp_format, cantools.logreader.TimestampFormat.ABSOLUTE)
        self.assertEqual(outp.timestamp.year, 2021)
        self.assertEqual(outp.timestamp.month, 2)
        self.assertEqual(outp.timestamp.day, 18)
        self.assertEqual(outp.timestamp.hour, 13)
        self.assertEqual(outp.timestamp.minute, 48)
        self.assertEqual(outp.timestamp.second, 24)
        self.assertEqual(outp.timestamp.microsecond, 501098)

    def test_candump_log_absolute_timestamp(self):
        parser = cantools.logreader.Parser()

        outp = parser.parse("(2020-12-19 12:04:45.485261)  vcan0  0C8   [8]  F0 00 00 00 00 00 00 00")
        self.assertEqual(outp.channel, 'vcan0')
        self.assertEqual(outp.frame_id, 0xc8)
        self.assertEqual(outp.data, b'\xF0\x00\x00\x00\x00\x00\x00\x00')
        self.assertEqual(outp.timestamp_format, cantools.logreader.TimestampFormat.ABSOLUTE)
        self.assertEqual(outp.timestamp.year, 2020)
        self.assertEqual(outp.timestamp.month, 12)
        self.assertEqual(outp.timestamp.day, 19)
        self.assertEqual(outp.timestamp.hour, 12)
        self.assertEqual(outp.timestamp.minute, 4)
        self.assertEqual(outp.timestamp.second, 45)
        self.assertEqual(outp.timestamp.microsecond, 485261)

    def test_candump_log_ascii(self):
        parser = cantools.logreader.Parser()

        outp = parser.parse(" can1  123   [8]  31 30 30 2E 35 20 46 4D   '100.5 FM'")
        self.assertEqual(outp.channel, 'can1')
        self.assertEqual(outp.frame_id, 0x123)
        self.assertEqual(outp.data, b'\x31\x30\x30\x2E\x35\x20\x46\x4D')
        self.assertEqual(outp.timestamp_format, cantools.logreader.TimestampFormat.MISSING)

    def test_candump_log_ascii_timestamped(self):
        parser = cantools.logreader.Parser()

        outp = parser.parse("  (1621271100.919019)  can1  123   [8]  31 30 30 2E 35 20 46 4D   '100.5 FM'")
        self.assertEqual(outp.channel, 'can1')
        self.assertEqual(outp.frame_id, 0x123)
        self.assertEqual(outp.data, b'\x31\x30\x30\x2E\x35\x20\x46\x4D')
        self.assertEqual(outp.timestamp_format, cantools.logreader.TimestampFormat.ABSOLUTE)

    def test_candump_log_ascii_absolute(self):
        parser = cantools.logreader.Parser()

        outp = parser.parse("(2020-12-19 12:04:45.485261)  can1  123   [8]  31 30 30 2E 35 20 46 4D   '100.5 FM'")
        self.assertEqual(outp.channel, 'can1')
        self.assertEqual(outp.frame_id, 0x123)
        self.assertEqual(outp.data, b'\x31\x30\x30\x2E\x35\x20\x46\x4D')
        self.assertEqual(outp.timestamp_format, cantools.logreader.TimestampFormat.ABSOLUTE)
        self.assertEqual(outp.timestamp.year, 2020)
        self.assertEqual(outp.timestamp.month, 12)
        self.assertEqual(outp.timestamp.day, 19)
        self.assertEqual(outp.timestamp.hour, 12)
        self.assertEqual(outp.timestamp.minute, 4)
        self.assertEqual(outp.timestamp.second, 45)
        self.assertEqual(outp.timestamp.microsecond, 485261)

    def test_pcan_traceV10(self):
        parser = cantools.logreader.Parser()

        outp = parser.parse("1) 1841 0001 8 F0 00 00 00 00 00 00 00")
        self.assertEqual(outp.channel, 'pcanx')
        self.assertEqual(outp.frame_id, 0x0001)
        self.assertEqual(outp.data, b'\xF0\x00\x00\x00\x00\x00\x00\x00')
        self.assertEqual(outp.timestamp_format, cantools.logreader.TimestampFormat.RELATIVE)
        self.assertEqual(outp.timestamp.seconds, 1)
        self.assertEqual(outp.timestamp.microseconds, 841000)

    def test_pcan_traceV11(self):
        parser = cantools.logreader.Parser()

        outp = parser.parse("1)      6357.2 Rx        0401  8    F0 00 00 00 00 00 00 00")
        self.assertEqual(outp.channel, 'pcanx')
        self.assertEqual(outp.frame_id, 0x0401)
        self.assertEqual(outp.data, b'\xF0\x00\x00\x00\x00\x00\x00\x00')
        self.assertEqual(outp.timestamp_format, cantools.logreader.TimestampFormat.RELATIVE)
        self.assertEqual(outp.timestamp.seconds, 6)
        self.assertEqual(outp.timestamp.microseconds, 357200)

    def test_pcan_traceV12(self):
        parser = cantools.logreader.Parser()

        outp = parser.parse("1)      6357.213 1  Rx        0401  8    F0 00 00 00 00 00 00 00")
        self.assertEqual(outp.channel, 'pcan1')
        self.assertEqual(outp.frame_id, 0x0401)
        self.assertEqual(outp.data, b'\xF0\x00\x00\x00\x00\x00\x00\x00')
        self.assertEqual(outp.timestamp_format, cantools.logreader.TimestampFormat.RELATIVE)
        self.assertEqual(outp.timestamp.seconds, 6)
        self.assertEqual(outp.timestamp.microseconds, 357213)

    def test_pcan_traceV13(self):
        parser = cantools.logreader.Parser()

        outp = parser.parse("1)      6357.213 1  Rx        0401 -  8    F0 00 00 00 00 00 00 00")
        self.assertEqual(outp.channel, 'pcan1')
        self.assertEqual(outp.frame_id, 0x0401)
        self.assertEqual(outp.data, b'\xF0\x00\x00\x00\x00\x00\x00\x00')
        self.assertEqual(outp.timestamp_format, cantools.logreader.TimestampFormat.RELATIVE)
        self.assertEqual(outp.timestamp.seconds, 6)
        self.assertEqual(outp.timestamp.microseconds, 357213)

    def test_pcan_traceV20(self):
        parser = cantools.logreader.Parser()

        outp = parser.parse(" 1      1059.900 DT 0300 Rx 7 00 00 00 00 04 00 00")
        self.assertEqual(outp.channel, 'pcanx')
        self.assertEqual(outp.frame_id, 0x0300)
        self.assertEqual(outp.data, b'\x00\x00\x00\x00\x04\x00\x00')
        self.assertEqual(outp.timestamp_format, cantools.logreader.TimestampFormat.RELATIVE)
        self.assertEqual(outp.timestamp.seconds, 1)
        self.assertEqual(outp.timestamp.microseconds, 59900)

    def test_pcan_traceV21(self):
        parser = cantools.logreader.Parser()

        outp = parser.parse(" 1      1059.900 DT 1 0300 Rx - 7 00 00 00 00 04 00 00")
        self.assertEqual(outp.channel, 'pcan1')
        self.assertEqual(outp.frame_id, 0x0300)
        self.assertEqual(outp.data, b'\x00\x00\x00\x00\x04\x00\x00')
        self.assertEqual(outp.timestamp_format, cantools.logreader.TimestampFormat.RELATIVE)
        self.assertEqual(outp.timestamp.seconds, 1)
        self.assertEqual(outp.timestamp.microseconds, 59900)

class TestLogreaderStreams(unittest.TestCase):
    def test_candump(self):
        testvec = io.StringIO("""\
  vcan0  0C8   [8]  F0 00 00 00 00 00 00 00
  vcan0  064   [10]  F0 01 FF FF FF FF FF FF FF FF
  vcan0  ERROR

  vcan0  1F4   [4]  01 02 03 04
  vcan0  1F3   [3]  01 02 03
""")
        parser = cantools.logreader.Parser(testvec)
        frame_iter = iter(parser)
        f1 = next(frame_iter)
        self.assertEqual(f1.frame_id, 0xc8)
        f2 = next(frame_iter)
        self.assertEqual(f2.frame_id, 0x64)
        f3 = next(frame_iter)
        self.assertEqual(f3.frame_id, 0x1f4)
        f4 = next(frame_iter)
        self.assertEqual(f4.frame_id, 0x1f3)

    def test_candump_absolute_timestamp(self):
        testvec = io.StringIO("""\
 (2020-12-19 12:04:45.485261)  vcan0  0C8   [8]  F0 00 00 00 00 00 00 00
 (2020-12-19 12:04:48.597222)  vcan0  064   [8]  F0 01 FF FF FF FF FF FF
 (2020-12-19 12:04:56.805087)  vcan0  1F4   [4]  01 02 03 04
 (2020-12-19 12:04:59.085517)  vcan0  1F3   [3]  01 02 03
""")
        parser = cantools.logreader.Parser(testvec)
        frame_iter = iter(parser)
        f1 = next(frame_iter)
        self.assertEqual(f1.frame_id, 0xc8)
        f2 = next(frame_iter)
        self.assertEqual(f2.frame_id, 0x64)
        f3 = next(frame_iter)
        self.assertEqual(f3.frame_id, 0x1f4)
        f4 = next(frame_iter)
        self.assertEqual(f4.frame_id, 0x1f3)

    def test_candump_time_since_start(self):
        testvec = io.StringIO("""\
 (000.000000)  vcan0  0C8   [8]  F0 00 00 00 00 00 00 00
 (002.047817)  vcan0  064   [8]  F0 01 FF FF FF FF FF FF
 (012.831664)  vcan0  1F4   [4]  01 02 03 04
 (015.679614)  vcan0  1F3   [3]  01 02 03
""")
        parser = cantools.logreader.Parser(testvec)
        frame_iter = iter(parser)
        f1 = next(frame_iter)
        self.assertEqual(f1.frame_id, 0xc8)
        f2 = next(frame_iter)
        self.assertEqual(f2.frame_id, 0x64)
        f3 = next(frame_iter)
        self.assertEqual(f3.frame_id, 0x1f4)
        f4 = next(frame_iter)
        self.assertEqual(f4.frame_id, 0x1f3)

    def test_candump_log_fd_absolute_time(self):
        testvec = io.StringIO("""\
  (1613656104.493702) can2 102##1150B7F0102010010000064A0020000100000000000E41F000000000090D1FF000020A600000000210100000000000000
""")
        parser = cantools.logreader.Parser(testvec)
        frame_iter = iter(parser)
        f1 = next(frame_iter)
        self.assertEqual(f1.frame_id, 0x102)

    def test_candump_log(self):
        testvec = io.StringIO("""\
(1594172461.968006) vcan0 0C8#F000000000000000
(1594172462.126542) vcan0 064#F001FFFFFFFFFFFFFFFF
(1594172462.127684) vcan0 ERROR

(1594172462.356874) vcan0 1F4#01020304
(1594172462.688432) vcan0 1F3#010203
""")
        parser = cantools.logreader.Parser(testvec)
        frame_iter = iter(parser)
        f1 = next(frame_iter)
        self.assertEqual(f1.frame_id, 0xc8)
        f2 = next(frame_iter)
        self.assertEqual(f2.frame_id, 0x64)
        f3 = next(frame_iter)
        self.assertEqual(f3.frame_id, 0x1f4)
        f4 = next(frame_iter)
        self.assertEqual(f4.frame_id, 0x1f3)

    def test_pcan_traceV10(self):
        testvec = io.StringIO("""\
;##########################################################################
; C:\\TraceFile.trc
;
; CAN activities recorded by PCAN Explorer
; Start time: 11.09.2002 16:00:20.682
; PCAN-Net: PCI1
;
; Columns description:
; ~~~~~~~~~~~~~~~~~~~~~
; +-current number in actual sample
; | +time offset of message (ms)
; | | +ID of message (hex)
; | | | +data length code
; | | | | +data bytes (hex) ...
; | | | | |
;----+- ---+--- ----+--- + -+ -- -- ...
1) 1841 0001 8 00 00 00 00 00 00 00 00
2) 1842 0008 4 ERROR 00 19 08 08
3) 1843 FFFFFFFF 4 00 00 00 04 -- -- -- -- BUSLIGHT
4) 1844 0100 3 RTR
5) 1845 0300 8 00 00 00 00 00 00 00 00
""")
        parser = cantools.logreader.Parser(testvec)
        frame_iter = iter(parser)
        f1 = next(frame_iter)
        self.assertEqual(f1.frame_id, 0x001)
        f2 = next(frame_iter)
        self.assertEqual(f2.frame_id, 0x300)

    def test_pcan_traceV11(self):
        testvec = io.StringIO("""\
;$FILEVERSION=1.1
;$STARTTIME=37704.5364870833
;
; C:\\TraceFile.trc
;
; Start time: 24.03.2003 12:52:32.484
; PCAN-Net: TestNet
;
; Columns description:
; ~~~~~~~~~~~~~~~~~~~~~
; +-Message Number
; | +Time Offset (ms)
; | | +Type
; | | | +ID (hex)
; | | | | +Data Length Code
; | | | | | +Data Bytes (hex) ...
; | | | | | |
;---+-- ----+---- --+-- ----+--- + -+ -- -- -- -- -- -- --
1) 1059.9 Rx 0300 7 00 00 00 00 04 00 00
2) 1283.2 Rx 0300 7 00 00 00 00 04 00 00
3) 1298.9 Tx 0400 2 00 00
4) 1323.0 Rx 0300 7 00 00 00 00 06 00 00
5) 1346.8 Warng FFFFFFFF 4 00 00 00 04 BUSLIGHT
6) 1349.2 Error 0008 4 00 19 08 08
7) 1352.7 Rx 0100 3 RTR
""")
        parser = cantools.logreader.Parser(testvec)
        frame_iter = iter(parser)
        f1 = next(frame_iter)
        self.assertEqual(f1.frame_id, 0x300)
        f2 = next(frame_iter)
        self.assertEqual(f2.frame_id, 0x300)
        f3 = next(frame_iter)
        self.assertEqual(f3.frame_id, 0x400)
        f4 = next(frame_iter)
        self.assertEqual(f4.frame_id, 0x300)
        f5 = next(frame_iter)
        self.assertEqual(f5.frame_id, 0x008)

    def test_pcan_traceV12(self):
        testvec = io.StringIO("""\
;$FILEVERSION=1.2
;$STARTTIME=39878.6772258947;
; C:\\TraceFile.trc
;
; Start time: 06.03.2009 16:15:12.317.3
; Connection: TestNet
;
; Columns description:
; ~~~~~~~~~~~~~~~~~~~~~
; +-Message Number
; | +Time Offset (ms)
; | | +Bus
; | | | +Type
; | | | | +ID (hex)
; | | | | | +Data Length Code
; | | | | | | +Data Bytes (hex) ...
; | | | | | | |
;---+-- ----+------ + --+-- ----+--- + -+ -- -- -- -- -- -- --
1) 1059.900 1 Rx 0300 7 00 00 00 00 04 00 00
2) 1283.231 1 Rx 0300 7 00 00 00 00 04 00 00
3) 1298.945 1 Tx 0400 2 00 00
4) 1323.201 1 Rx 0300 7 00 00 00 00 06 00 00
5) 1346.834 1 Warng FFFFFFFF 4 00 00 00 04 BUSLIGHT
6) 1349.222 1 Error 0008 4 00 19 08 08
7) 1352.743 1 Rx 0100 3 RTR
""")
        parser = cantools.logreader.Parser(testvec)
        frame_iter = iter(parser)
        f1 = next(frame_iter)
        self.assertEqual(f1.frame_id, 0x300)
        f2 = next(frame_iter)
        self.assertEqual(f2.frame_id, 0x300)
        f3 = next(frame_iter)
        self.assertEqual(f3.frame_id, 0x400)
        f4 = next(frame_iter)
        self.assertEqual(f4.frame_id, 0x300)
        f5 = next(frame_iter)
        self.assertEqual(f5.frame_id, 0x008)

    def test_pcan_traceV13(self):
        testvec = io.StringIO("""\
;$FILEVERSION=1.3
;$STARTTIME=44656.4982577909614
;
;   Start time: 05/04/2022 11:57:29.473.1
;   Generated by PCAN-Explorer v6.5.2.2576
;-------------------------------------------------------------------------------
;   Bus  Connection   Net Connection             Protocol  Bit rate
;   1    Connection1  PCANLight_USB_16@pcan_usb  CAN       250 kbit/s
;-------------------------------------------------------------------------------
;   Message   Time    Bus  Type   ID    Reserved
;   Number    Offset  |    |      [hex] |   Data Length Code
;   |         [ms]    |    |      |     |   |    Data [hex] ...
;   |         |       |    |      |     |   |    |
;---+-- ------+------ +- --+-- ---+---- +- -+-- -+ -- -- -- -- -- -- --
     1)      6357.213 1  Rx        0401 -  8    00 00 00 00 00 00 00 00
     2)      6357.682 1  Rx        04A1 -  8    00 00 00 00 00 00 00 00
     3)      1346.834 1  Warng FFFFFFFF -  4    00 00 00 04 BUSLIGHT
     4)      6377.180 1  Rx        0400 -  8    00 00 00 00 00 00 00 00
     5)      6377.692 1  Rx        04A0 -  8    00 00 00 00 00 00 00 00
""")
        parser = cantools.logreader.Parser(testvec)
        frame_iter = iter(parser)
        f1 = next(frame_iter)
        self.assertEqual(f1.frame_id, 0x401)
        f2 = next(frame_iter)
        self.assertEqual(f2.frame_id, 0x4A1)
        f3 = next(frame_iter)
        self.assertEqual(f3.frame_id, 0x400)
        f4 = next(frame_iter)
        self.assertEqual(f4.frame_id, 0x4A0)

    def test_pcan_traceV20(self):
        testvec = io.StringIO("""\
;$FILEVERSION=2.0
;$STARTTIME=42209.4075997106
;$COLUMNS=N,O,T,I,d,l,D
;
; C:\\TraceFile.trc
; Start time: 24.07.2015 09:46:56.615.0
; Generated by PCAN-View v4.0.29.426
;-------------------------------------------------------------------------------
; Connection Bit rate
; PCANLight_USB_16@pcan_usb Nominal 1 MBit/s, Data 2 Mbit/s
;-------------------------------------------------------------------------------
; Message Time Type ID Rx/Tx
; Number Offset | [hex] | Data Length
; | [ms] | | | | Data [hex] ...
; | | | | | | |
;---+-- ------+------ +- --+----- +- +- +- -- -- -- -- -- -- --
1 1059.900 DT 0300 Rx 7 00 00 00 00 04 00 00
2 1283.231 DT 0300 Rx 7 00 00 00 00 04 00 00
3 1298.945 DT 0400 Tx 2 00 00
4 1323.201 DT 0300 Rx 7 00 00 00 00 06 00 00
5 1334.416 FD 0500 Tx 12 01 02 03 04 05 06 07 08 09 0A 0B 0C
6 1334.522 ER Rx 04 00 02 00 00
7 1334.531 ST Rx 00 00 00 08
8 1334.643 EC Rx 02 02
9 1335.156 DT 18EFC034 Tx 8 01 02 03 04 05 06 07 08
10 1336.543 RR 0100 Rx 3
""")
        parser = cantools.logreader.Parser(testvec)
        frame_iter = iter(parser)
        f1 = next(frame_iter)
        self.assertEqual(f1.frame_id, 0x300)
        f2 = next(frame_iter)
        self.assertEqual(f2.frame_id, 0x300)
        f3 = next(frame_iter)
        self.assertEqual(f3.frame_id, 0x400)
        f4 = next(frame_iter)
        self.assertEqual(f4.frame_id, 0x300)
        f5 = next(frame_iter)
        self.assertEqual(f5.frame_id, 0x500)
        f6 = next(frame_iter)
        self.assertEqual(f6.frame_id, 0x18EFC034)

    def test_pcan_traceV21(self):
        testvec = io.StringIO("""\
;$FILEVERSION=2.1
;$STARTTIME=41766.4648963872
;$COLUMNS=N,O,T,B,I,d,R,L,D
;
; C:\\TraceFile.trc
; Start time: 07.05.2015 11:09:27.047.8
; Generated by PCAN-Explorer v6.0.0
;-------------------------------------------------------------------------------
; Bus Name Connection Protocol
; 1 Connection1 TestNet@pcan_usb CAN
;-------------------------------------------------------------------------------
; Message Time Type ID Rx/Tx
; Number Offset | Bus [hex] | Reserved
; | [ms] | | | | | Data Length Code
; | | | | | | | | Data [hex] ...
; | | | | | | | | |
;---+-- ------+------ +- +- --+----- +- +- +--- +- -- -- -- -- -- -- --
1 1059.900 DT 1 0300 Rx - 7 00 00 00 00 04 00 00
2 1283.231 DT 1 0300 Rx - 7 00 00 00 00 04 00 00
3 1298.945 DT 1 0400 Tx - 2 00 00
4 1323.201 DT 1 0300 Rx - 7 00 00 00 00 06 00 00
5 1334.416 FD 1 0500 Tx - 12 01 02 03 04 05 06 07 08 09 0A 0B 0C
6 1334.222 ER 1 - Rx - 5 04 00 02 00 00
7 1334.224 EV 1 User-defined event for bus 1
8 1334.225 EV - User-defined event for all busses
9 1334.231 ST 1 - Rx - 4 00 00 00 08
10 1334.268 ER 1 - Rx - 5 04 00 02 08 00
11 1334.643 EC 1 - Rx - 2 02 02
12 1335.156 DT 1 18EFC034 Tx - 8 01 02 03 04 05 06 07 08
13 1336.543 RR 1 0100 Rx - 3
""")
        parser = cantools.logreader.Parser(testvec)
        frame_iter = iter(parser)
        f1 = next(frame_iter)
        self.assertEqual(f1.frame_id, 0x300)
        f2 = next(frame_iter)
        self.assertEqual(f2.frame_id, 0x300)
        f3 = next(frame_iter)
        self.assertEqual(f3.frame_id, 0x400)
        f4 = next(frame_iter)
        self.assertEqual(f4.frame_id, 0x300)
        f5 = next(frame_iter)
        self.assertEqual(f5.frame_id, 0x500)
        f6 = next(frame_iter)
        self.assertEqual(f6.frame_id, 0x18EFC034)



from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Literal,
    Mapping,
    NamedTuple,
    Optional,
    OrderedDict,
    Sequence,
    Tuple,
    Union,
)

from bitstruct import CompiledFormatDict
from typing_extensions import TypedDict

if TYPE_CHECKING:
    import os

    from .database import Message, Signal
    from .database.namedsignalvalue import NamedSignalValue


class Formats(NamedTuple):
    big_endian: CompiledFormatDict
    little_endian: CompiledFormatDict
    padding_mask: int


StringPathLike = Union[str, "os.PathLike[str]"]
Comments = Dict[Optional[str], str]
class Codec(TypedDict):
    signals: List["Signal"]
    formats: Formats
    multiplexers: Mapping[str, Mapping[int, Any]]

ByteOrder = Literal["little_endian", "big_endian"]
Choices = OrderedDict[int, Union[str, "NamedSignalValue"]]

# Type aliases. Introduced to reduce type annotation complexity while
# allowing for more complex encode/decode schemes like the one used
# for AUTOSAR container messages.
SignalValueType = Union[int, float, str, "NamedSignalValue"]
SignalDictType = Dict[str, SignalValueType]
SignalMappingType = Mapping[str, SignalValueType]
ContainerHeaderSpecType = Union["Message", str, int]
ContainerUnpackResultType = Sequence[Union[Tuple["Message", bytes], Tuple[int, bytes]]]
ContainerUnpackListType = List[Union[Tuple["Message", bytes], Tuple[int, bytes]]]
ContainerDecodeResultType = Sequence[
    Union[Tuple["Message", SignalMappingType], Tuple[int, bytes]]
]
ContainerDecodeResultListType = List[
    Union[Tuple["Message", SignalDictType], Tuple[int, bytes]]
]
ContainerEncodeInputType = Sequence[
    Tuple[ContainerHeaderSpecType, Union[bytes, SignalMappingType]]
]
DecodeResultType = Union[SignalDictType, ContainerDecodeResultType]
EncodeInputType = Union[SignalMappingType, ContainerEncodeInputType]

SecOCAuthenticatorFn = Callable[["Message", bytes, int], bytes]



# The tester module.

import queue
import time
from collections import UserDict
from typing import Dict, List, Optional

import can

from .errors import Error


class DecodedMessage:
    """A decoded message.

    """

    def __init__(self, name, signals):
        self.name = name
        self.signals = signals


class Messages(UserDict):
    def __setitem__(self, message_name, value):
        if getattr(self, '_frozen', False):
            if message_name not in self.data:
                raise KeyError(message_name)
        self.data[message_name] = value

    def __missing__(self, key):
        raise Error(f"invalid message name '{key}'")


def _invert_signal_tree(
        tree :List,
        cur_mpx: Optional[Dict] = None,
        ret: Optional[Dict] = None
) -> Dict:

    """The tree is laid out with two kinds of dicts.  Single-element dict
    keyed by string -> multiplexer, which is own dict keyed by
    integers.

    """

    if ret is None:
        ret = {}

    if cur_mpx is None:
        cur_mpx = {}

    for sigs in tree:
        if type(sigs) == dict: # outer signal keyed by muliplexer name
            (mpx_name, mpx_vals), = sigs.items()
            for mpx_val, sig_tree in mpx_vals.items(): # inner signal
                #                          keyed by multiplexer values
                next_mpx = cur_mpx.copy()
                next_mpx[mpx_name] = mpx_val
                _invert_signal_tree(sig_tree, next_mpx, ret)

        elif type(sigs) == str:
            ret.setdefault(sigs,[]).append(set(cur_mpx.items()))
        else:
            raise TypeError(repr(sigs))

    return ret

def invert_signal_tree(tree: List) -> Dict:
    """Return a mapping of signals to the multiplex settings that will
    yield the signal.

    {signal: [{mplexers}, {mplexers...}]}

    """
    return _invert_signal_tree(tree)

class Listener(can.Listener):

    def __init__(self, database, messages, input_queue, on_message):
        self._database = database
        self._messages = messages
        self._input_queue = input_queue
        self._on_message = on_message

    def on_message_received(self, msg):
        if msg.is_error_frame or msg.is_remote_frame:
            return

        try:
            database_message = self._database.get_message_by_frame_id(
                msg.arbitration_id)
        except KeyError:
            return

        if database_message.name not in self._messages:
            return

        message = self._messages[database_message.name]

        if not message.enabled:
            return

        decoded = DecodedMessage(database_message.name,
                                 database_message.decode(msg.data,
                                                         message.decode_choices,
                                                         message.scaling))

        if self._on_message:
            self._on_message(decoded)

        self._input_queue.put(decoded)


class Message(UserDict):

    def __init__(self,
                 database,
                 can_bus,
                 input_list,
                 input_queue,
                 decode_choices,
                 scaling,
                 padding):
        super().__init__()
        self.database = database
        self._mplex_map = invert_signal_tree(database.signal_tree)
        self._can_bus = can_bus
        self._input_queue = input_queue
        self.decode_choices = decode_choices
        self.scaling = scaling
        self.padding = padding
        self._input_list = input_list
        self.enabled = True
        self._can_message = None
        self._periodic_task = None
        self._signal_names = {s.name for s in self.database.signals}
        self.update(self._prepare_initial_signal_values())

    @property
    def periodic(self):
        return self.database.cycle_time is not None

    def __getitem__(self, signal_name):
        return self.data[signal_name]

    def __setitem__(self, signal_name, value):
        if signal_name not in self._signal_names:
            raise KeyError(signal_name)
        self.data[signal_name] = value
        self._update_can_message()

    def update(self, signals):
        s = dict(signals)
        new_signal_names = set(s) - self._signal_names
        if new_signal_names:
            raise KeyError(repr(new_signal_names))

        self.data.update(s)
        self._update_can_message()

    def send(self, signals=None):
        if signals is not None:
            self.update(signals)

        self._can_bus.send(self._can_message)

    def expect(self, signals=None, timeout=None, discard_other_messages=True):
        if signals is None:
            signals = {}

        decoded = self._expect_input_list(signals, discard_other_messages)

        if decoded is None:
            decoded = self._expect_input_queue(signals,
                                               timeout,
                                               discard_other_messages)

        return decoded

    def _expect_input_list(self, signals, discard_other_messages):
        other_messages = []

        while len(self._input_list) > 0:
            message = self._input_list.pop(0)
            decoded = self._filter_expected_message(message, signals)

            if decoded is not None:
                break

            other_messages.append(message)
        else:
            decoded = None

        if not discard_other_messages:
            other_messages += self._input_list
            del self._input_list[:]
            self._input_list.extend(other_messages)

        return decoded

    def _expect_input_queue(self, signals, timeout, discard_other_messages):
        if timeout is not None:
            end_time = time.time() + timeout
            remaining_time = timeout
        else:
            remaining_time = None

        while True:
            try:
                message = self._input_queue.get(timeout=remaining_time)
            except queue.Empty:
                return

            decoded = self._filter_expected_message(message, signals)

            if decoded is not None:
                return decoded

            if not discard_other_messages:
                self._input_list.append(message)

            if timeout is not None:
                remaining_time = end_time - time.time()

                if remaining_time <= 0:
                    return

    def _filter_expected_message(self, message, signals):
        if message.name == self.database.name:
            if all(message.signals[name] == signals[name] for name in signals):
                return message.signals

    def send_periodic_start(self):
        if not self.enabled:
            return

        self._periodic_task = self._can_bus.send_periodic(
            self._can_message,
            self.database.cycle_time / 1000.0)

    def send_periodic_stop(self):
        if self._periodic_task is not None:
            self._periodic_task.stop()
            self._periodic_task = None

    def _update_can_message(self):
        arbitration_id = self.database.frame_id
        extended_id = self.database.is_extended_frame
        pruned_data = self.database.gather_signals(self.data)
        data = self.database.encode(pruned_data,
                                    self.scaling,
                                    self.padding)
        self._can_message = can.Message(arbitration_id=arbitration_id,
                                        is_extended_id=extended_id,
                                        data=data)

        if self._periodic_task is not None:
            self._periodic_task.modify_data(self._can_message)

    def _prepare_initial_signal_values(self):
        initial_sig_values = {}

        # Choose a valid set of mux settings
        mplex_settings = {}
        for m0 in reversed(self._mplex_map.values()):
            for m1 in m0:
                mplex_settings.update(m1)

        for signal in self.database.signals:
            minimum = 0 if not signal.minimum else signal.minimum
            maximum = 0 if not signal.maximum else signal.maximum
            if signal.initial:
                # use initial signal value (if set)
                initial_sig_values[signal.name] = signal.initial
            elif signal.is_multiplexer:
                initial_sig_values[signal.name] = mplex_settings.get(signal.name, 0)
            elif minimum <= 0 <= maximum:
                # use 0 if in allowed range
                initial_sig_values[signal.name] = 0
            else:
                # set at least some default value
                initial_sig_values[signal.name] = minimum
        return initial_sig_values


class Tester:
    """Test given node `dut_name` on given CAN bus `bus_name`.

    `database` is a :class:`~cantools.database.can.Database` instance.

    `can_bus` a CAN bus object, normally created using the python-can
    package.

    The `on_message` callback is called for every successfully decoded
    received message. It is called with one argument, an
    :class:`~cantools.tester.DecodedMessage` instance.

    Here is an example of how to create a tester:

    >>> import can
    >>> import cantools
    >>> can.rc['interface'] = 'socketcan'
    >>> can.rc['channel'] = 'vcan0'
    >>> can_bus = can.interface.Bus()
    >>> database = cantools.database.load_file('tests/files/tester.kcd')
    >>> tester = cantools.tester.Tester('PeriodicConsumer', database, can_bus, 'PeriodicBus')

    """

    def __init__(self,
                 dut_name,
                 database,
                 can_bus,
                 bus_name=None,
                 on_message=None,
                 decode_choices=True,
                 scaling=True,
                 padding=False):
        self._dut_name = dut_name
        self._bus_name = bus_name
        self._database = database
        self._can_bus = can_bus
        self._input_list = []
        self._input_queue = queue.Queue()
        self._messages = Messages()
        self._is_running = False

        # DUT name validation.
        node_names = [node.name for node in database.nodes]

        if dut_name and not any(name == dut_name for name in node_names):
            raise Error(
                "expected DUT name in {}, but got '{}'".format(node_names,
                                                               dut_name))

        # BUS name validation.
        bus_names = [bus.name for bus in database.buses]

        if len(bus_names) == 0:
            if bus_name is not None:
                raise Error(
                    "expected bus name None as there are no buses defined in "
                    "the database, but got '{}'".format(bus_name))
        elif not any(name == bus_name for name in bus_names):
            raise Error(
                "expected bus name in {}, but got '{}'".format(bus_names,
                                                               bus_name))

        for message in database.messages:
            if message.bus_name == bus_name:
                self._messages[message.name] = Message(message,
                                                       can_bus,
                                                       self._input_list,
                                                       self._input_queue,
                                                       decode_choices,
                                                       scaling,
                                                       padding)

        listener = Listener(self._database,
                            self._messages,
                            self._input_queue,
                            on_message)
        self._notifier = can.Notifier(can_bus, [listener])
        self._messages._frozen = True

    def start(self):
        """Start the tester. Starts sending enabled periodic messages.

        >>> tester.start()

        """

        for message in self._messages.values():
            if self._dut_name and self._dut_name in message.database.senders:
                continue

            if not message.periodic:
                continue

            message.send_periodic_start()

        self._is_running = True

    def stop(self):
        """Stop the tester. Periodic messages will not be sent after this
        call. Call :meth:`~cantools.tester.Tester.start()` to resume a
        stopped tester.

        >>> tester.stop()

        """

        for message in self._messages.values():
            message.send_periodic_stop()

        self._is_running = False

    @property
    def messages(self):
        """Set and get signals in messages. Set signals takes effect
        immediately for started enabled periodic messages. Call
        :meth:`~cantools.tester.Tester.send()` for other messages.

        >>> periodic_message = tester.messages['PeriodicMessage1']
        >>> periodic_message
        {'Signal1': 0, 'Signal2': 0}
        >>> periodic_message['Signal1'] = 1
        >>> periodic_message.update({'Signal1': 2, 'Signal2': 5})
        >>> periodic_message
        {'Signal1': 2, 'Signal2': 5}

        """

        return self._messages

    def enable(self, message_name):
        """Enable given message `message_name` and start sending it if its
        periodic and the tester is running.

        >>> tester.enable('PeriodicMessage1')

        """

        message = self._messages[message_name]
        message.enabled = True

        if self._is_running and message.periodic:
            message.send_periodic_start()

    def disable(self, message_name):
        """Disable given message `message_name` and stop sending it if its
        periodic, enabled and the tester is running.

        >>> tester.disable('PeriodicMessage1')

        """

        message = self._messages[message_name]
        message.enabled = False

        if self._is_running and message.periodic:
            message.send_periodic_stop()

    def send(self, message_name, signals=None):
        """Send given message `message_name` and optional signals `signals`.

        >>> tester.send('Message1', {'Signal2': 10})
        >>> tester.send('Message1')

        """

        self._messages[message_name].send(signals)

    def expect(self,
               message_name,
               signals=None,
               timeout=None,
               discard_other_messages=True):
        """Expect given message `message_name` and signal values `signals`
        within `timeout` seconds.

        Give `signals` as ``None`` to expect any signal values.

        Give `timeout` as ``None`` to wait forever.

        Messages are read from the input queue, and those not matching
        given `message_name` and `signals` are discarded if
        `discard_other_messages` is
        ``True``. :meth:`~cantools.tester.Tester.flush_input()` may be
        called to discard all old messages in the input queue before
        calling the expect function.

        Returns the expected message, or ``None`` on timeout.

        >>> tester.expect('Message2', {'Signal1': 13})
        {'Signal1': 13, 'Signal2': 9}

        """

        return self._messages[message_name].expect(signals,
                                                   timeout,
                                                   discard_other_messages)

    def flush_input(self):
        """Flush, or discard, all messages in the input queue.

        """

        del self._input_list[:]

        while not self._input_queue.empty():
            self._input_queue.get()



# -*- coding: utf-8 -*-

from parameterized import parameterized
import unittest

from cantools.database.utils import sawtooth_to_network_bitnum, cdd_offset_to_dbc_start_bit

class CanToolsDatabaseUtilsTest(unittest.TestCase):

    @parameterized.expand((
        ("0", 0, 7),
        ("1", 1, 6),
        ("2", 2, 5),
        ("3", 3, 4),
        ("4", 4, 3),
        ("5", 5, 2),
        ("6", 6, 1),
        ("7", 7, 0),
        ("8", 8, 15),
        ("15", 15, 8),
        ("32", 32, 39),
        ("64", 64, 71),
        ("65", 65, 70),
    ))
    def test_sawtooth_to_network_bitnum(self, _name, sawtooth_bitnum, expected_network_bitnum):

        network_bitnum = sawtooth_to_network_bitnum(sawtooth_bitnum)
        self.assertEqual(network_bitnum, expected_network_bitnum)

    @parameterized.expand((
        ("BE-0-8", "big_endian", 0, 8, 7),
        ("LE-0-8", "little_endian", 0, 8, 0),

        ("BE-0-4", "big_endian", 0, 4, 3),
        ("LE-0-4", "little_endian", 0, 4, 0),

        ("BE-4-4", "big_endian", 4, 4, 7),
        ("LE-4-4", "little_endian", 4, 4, 4),

        ("BE-0-16", "big_endian", 0, 16, 7),
        ("LE-0-16", "little_endian", 0, 16, 0),

        ("BE-0-32", "big_endian", 0, 32, 7),
        ("LE-0-32", "little_endian", 0, 32, 0),

        ("BE-32-16", "big_endian", 32, 16, 39),
        ("LE-32-16", "little_endian", 32, 16, 32),
    ))
    def test_cdd_offset_to_dbc_start_bit(self, _name, byte_order, offset, bit_length, expected_dbc_start_bit):

        dbc_start_bit = cdd_offset_to_dbc_start_bit(offset, bit_length, byte_order)
        self.assertEqual(dbc_start_bit, expected_dbc_start_bit)


if __name__ == '__main__':
    unittest.main()



# unit tests for the autosar specifics of message and database objects
# (message.autosar and db.autosar)
import unittest
import traceback

import cantools
import cantools.autosar

from cantools.autosar.snakeauth import SnakeOilAuthenticator

class CanToolsAutosarTest(unittest.TestCase):
    def test_autosar3_e2e_profile2(self):
        db = cantools.db.load_file('tests/files/arxml/system-3.2.3.arxml')
        msg = db.get_message_by_name('Status')

        # verify the parameters
        self.assertIsNotNone(msg.autosar)
        self.assertIsNotNone(msg.autosar.e2e)
        self.assertEqual(msg.autosar.e2e.category, 'p2')
        self.assertEqual(msg.autosar.e2e.data_ids, list(range(0,16)))

        # test the crc calculation
        did = msg.autosar.e2e.data_ids
        self.assertEqual(cantools.autosar.compute_profile2_crc(b'\x00\x51\x22',
                                                               did[5]),
                         173)
        self.assertEqual(cantools.autosar.compute_profile2_crc(b'\x00\x51\x22',
                                                               msg),
                         17)


        # test the apply function
        self.assertIsNone(cantools.autosar.apply_profile2_crc(b'\xff',
                                                              msg))
        self.assertEqual(cantools.autosar.apply_profile2_crc(b'\xff\x51\x22',
                                                             msg),
                         b'\x11\x51\x22')

        # test the check function
        self.assertFalse(cantools.autosar.check_profile2_crc(b'\xff',
                                                             msg))
        self.assertFalse(cantools.autosar.check_profile2_crc(b'\x00\x51\x22',
                                                             msg))
        self.assertTrue(cantools.autosar.check_profile2_crc(b'\x11\x51\x22',
                                                            msg))

        # make sure that profile 5 protected messages cannot be used
        # with the functions for profile 2
        db2 = cantools.db.load_file('tests/files/arxml/system-4.2.arxml')
        pr5msg = db2.get_message_by_name('Message3')
        self.assertIsNone(cantools.autosar.apply_profile2_crc(b'\xff\xff',
                                                              pr5msg))
        self.assertIsNone(cantools.autosar.check_profile2_crc(b'\xff\xff',
                                                              pr5msg))

    def test_autosar4_e2e_profile5(self):
        db = cantools.db.load_file('tests/files/arxml/system-4.2.arxml')
        msg = db.get_message_by_name('Message3')

        # verify the parameters
        self.assertIsNotNone(msg.autosar)
        self.assertTrue(msg.autosar.is_secured)
        self.assertEqual(msg.autosar.secoc.payload_length, 4)
        self.assertIsNotNone(msg.autosar.e2e)
        self.assertEqual(msg.autosar.e2e.category, 'Profile5')
        self.assertEqual(msg.autosar.e2e.data_ids, [321])

        # test the crc calculation
        self.assertEqual(cantools.autosar.compute_profile5_crc(
            b'\x00\x00\x11\x22', msg.autosar.e2e.data_ids[0]),
                         12201)
        self.assertEqual(cantools.autosar.compute_profile5_crc(
            b'\xff\xff\x11\x22\x33\x44\x55', msg),
                         12201
                         )

        # test  the apply function
        self.assertEqual(cantools.autosar.apply_profile5_crc(b'\xff\x51\x22',
                                                             msg),
                         None)
        self.assertEqual(cantools.autosar.apply_profile5_crc(b'\xff\xee\x22\x33',
                                                             msg),
                         b'\xcf\xec"3')

        # test the check function
        self.assertTrue(cantools.autosar.check_profile5_crc(b'\xcf\xec"3',
                                                             msg))
        self.assertFalse(cantools.autosar.check_profile5_crc(b'\xcf\xec"4',
                                                             msg))
        self.assertFalse(cantools.autosar.check_profile5_crc(b'\xcf\xec"',
                                                             msg))

        # make sure that profile 2 protected messages cannot be used
        # with the functions for profile 5
        db2 = cantools.db.load_file('tests/files/arxml/system-3.2.3.arxml')
        pr2msg = db2.get_message_by_name('Status')
        self.assertIsNone(
            cantools.autosar.apply_profile5_crc(b'\xff\xff\xff\xff', pr2msg))
        self.assertIsNone(
            cantools.autosar.check_profile5_crc(b'\xff\xff\xff\xff', pr2msg))

    def test_autosar4_e2e(self):
        db = cantools.db.load_file('tests/files/arxml/system-4.2.arxml')

        # verify the E2E parameters for a message that uses profile 5
        msg = db.get_message_by_name('Message3')
        self.assertTrue(msg.autosar is not None)
        self.assertTrue(msg.autosar.e2e is not None)
        self.assertEqual(msg.autosar.e2e.category, 'Profile5')
        self.assertEqual(msg.autosar.e2e.data_ids, [321])

        # verify the E2E parameters for a message that uses profile 2
        msg = db.get_message_by_name('Message1')
        self.assertTrue(msg.autosar is not None)
        self.assertTrue(msg.autosar.e2e is not None)
        self.assertEqual(msg.autosar.e2e.category, 'Profile2')
        self.assertEqual(msg.autosar.e2e.data_ids, list(range(123, 123+16)))

        # verify the E2E parameters for a contained message
        cmsg = db.get_message_by_name('OneToContainThemAll')
        msg = cmsg.get_contained_message_by_name('message1')
        self.assertTrue(msg.autosar is not None)
        self.assertFalse(msg.autosar.is_secured)
        self.assertTrue(msg.autosar.e2e is not None)
        self.assertEqual(msg.autosar.e2e.category, 'Profile2')
        self.assertEqual(msg.autosar.e2e.data_ids, list(range(123, 123+16)))
        self.assertEqual(msg.autosar.e2e.payload_length, msg.length)

        # verify the E2E parameters for a contained secured message
        cmsg = db.get_message_by_name('OneToContainThemAll')
        msg = cmsg.get_contained_message_by_name('message3_secured')
        self.assertTrue(msg.autosar is not None)
        self.assertTrue(msg.autosar.is_secured)
        self.assertTrue(msg.autosar.e2e is not None)
        self.assertEqual(msg.autosar.e2e.category, 'Profile5')
        self.assertEqual(msg.autosar.e2e.data_ids, [321])
        self.assertEqual(msg.autosar.e2e.payload_length, 4)

    def test_autosar4_secoc(self):
        db = cantools.db.load_file('tests/files/arxml/system-4.2.arxml')

        dbmsg = db.get_message_by_name('Message3')

        self.assertTrue(dbmsg.autosar is not None)
        self.assertTrue(dbmsg.autosar.is_secured)
        self.assertTrue(dbmsg.autosar.secoc is not None)
        self.assertEqual(dbmsg.autosar.secoc.freshness_algorithm_name,
                         'SmellyCheese')
        self.assertEqual(dbmsg.autosar.secoc.auth_algorithm_name,
                         'KnockKnock')
        self.assertEqual(dbmsg.autosar.secoc.payload_length, 4)
        self.assertEqual(dbmsg.autosar.secoc.data_id, 1337)
        self.assertEqual(dbmsg.autosar.secoc.freshness_bit_length, 32)
        self.assertEqual(dbmsg.autosar.secoc.freshness_tx_bit_length, 6)
        self.assertEqual(dbmsg.autosar.secoc.auth_tx_bit_length, 10)

        encoded_raw = bytes([0x00]*dbmsg.length)
        snake_auth = SnakeOilAuthenticator(secret="Psst! Top secretion!")
        encoded = cantools.autosar.apply_authenticator(encoded_raw,
                                                       dbmsg,
                                                       snake_auth,
                                                       0xcccc)

        self.assertEqual(encoded, bytes.fromhex('000000003130'))

        decoded = dbmsg.decode(encoded)
        self.assertEqual(decoded['Message3_Freshness'], 0xcccc&0x3f)
        self.assertEqual(decoded['Message3_Authenticator'], 304)

        self.assertTrue(cantools.autosar.verify_authenticator(encoded,
                                                              dbmsg,
                                                              snake_auth,
                                                              0xcccc))
        self.assertFalse(cantools.autosar.verify_authenticator(encoded,
                                                               dbmsg,
                                                               snake_auth,
                                                               0xdccc))

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3

import os
import sys
import unittest
from unittest import mock
from io import StringIO
import cantools
import matplotlib.pyplot as plt


class CanToolsPlotTest(unittest.TestCase):

    DBC_FILE = os.path.join(os.path.split(__file__)[0], 'files/dbc/abs.dbc')
    FN_OUT = "out.pdf"

    def test_plot_tz(self):
        self.assertFalse(os.path.exists(self.FN_OUT))
        argv = ['cantools', 'plot', '-o', self.FN_OUT, self.DBC_FILE]
        input_data = """\
 (000.000000)  vcan0  00000343   [8]  C5 04 B7 04 9B 04 C5 04
 (001.001787)  vcan0  00000343   [8]  69 04 69 04 77 04 7E 04
 (002.003592)  vcan0  00000343   [8]  29 04 30 04 29 04 22 04
 (003.005400)  vcan0  00000343   [8]  FC 03 20 04 20 04 FC 03
 (004.006942)  vcan0  00000343   [8]  DE 03 D0 03 D0 03 C9 03
 (005.008400)  vcan0  00000343   [8]  7E 03 85 03 8C 03 77 03
 (006.009926)  vcan0  00000343   [8]  65 03 3B 03 50 03 65 03
 (007.011457)  vcan0  00000343   [8]  17 03 3B 03 34 03 10 03
 (008.013215)  vcan0  00000343   [8]  00 03 F2 02 15 03 F9 02
 (009.014779)  vcan0  00000343   [8]  CB 02 BC 02 B5 02 D2 02
"""

        with mock.patch('sys.stdin', StringIO(input_data)):
            with mock.patch('sys.argv', argv):
                cantools._main()

        self.assertTrue(os.path.exists(self.FN_OUT))
        os.remove(self.FN_OUT)

    def test_plot_style(self):
        self.assertFalse(os.path.exists(self.FN_OUT))
        argv = ['cantools', 'plot', '--style', 'seaborn-v0_8-muted', '-o', self.FN_OUT, self.DBC_FILE]
        input_data = """\
 (000.000000)  vcan0  00000343   [8]  C5 04 B7 04 9B 04 C5 04
 (001.001787)  vcan0  00000343   [8]  69 04 69 04 77 04 7E 04
 (002.003592)  vcan0  00000343   [8]  29 04 30 04 29 04 22 04
 (003.005400)  vcan0  00000343   [8]  FC 03 20 04 20 04 FC 03
 (004.006942)  vcan0  00000343   [8]  DE 03 D0 03 D0 03 C9 03
 (005.008400)  vcan0  00000343   [8]  7E 03 85 03 8C 03 77 03
 (006.009926)  vcan0  00000343   [8]  65 03 3B 03 50 03 65 03
 (007.011457)  vcan0  00000343   [8]  17 03 3B 03 34 03 10 03
 (008.013215)  vcan0  00000343   [8]  00 03 F2 02 15 03 F9 02
 (009.014779)  vcan0  00000343   [8]  CB 02 BC 02 B5 02 D2 02
"""

        with mock.patch('sys.stdin', StringIO(input_data)):
            with mock.patch('sys.argv', argv):
                cantools._main()

        self.assertTrue(os.path.exists(self.FN_OUT))
        os.remove(self.FN_OUT)

    def test_plot_list_styles(self):
        self.assertFalse(os.path.exists(self.FN_OUT))
        argv = ['cantools', 'plot', '--list-styles', '']
        stdout = StringIO()

        expected = "available matplotlib styles:"
        expected += "".join("\n- %s" % s for s in plt.style.available)
        expected += "\n"

        with mock.patch('sys.stdout', stdout):
            with mock.patch('sys.argv', argv):
                cantools._main()

        self.assertEqual(stdout.getvalue(), expected)

if __name__ == '__main__':
    unittest.main()
